package rootgui.model;

import java.io.Serializable;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import rootgui.model.objects.SimulationModel;
import rootgui.model.objects.Species;
import rootgui.model.recordstore.IRecordStore;
import rootgui.model.recordstore.Record;
import rootgui.model.recordstore.RecordStore;
import rootgui.model.recordstore.RecordStoreActionListener;
import rootgui.model.recordstore.SpeciesRecord;

/*
 * Class that contains the data regarding one simulation
 */
public class SimulationRun implements IHistDataProvider, Serializable {

	/**
	 * 
	 */
	private static final long serialVersionUID = 777255007528922493L;

	protected boolean tissueFileInitialised = false;

	protected boolean simulationInitialised = false;

	protected Boolean simInProgress = false;

	/*
	 * 
	 */
	protected String tissueFileName;

	/*
	 * 
	 */
	protected Map<String, SimulationModel> loadedSimulationModels, availableSimulationModels;

	/*
	 * 
	 */
	protected double timestep;

	/*
	 * Number of iterations the simulation should run.
	 */
	protected int maxIterations;

	/*
	 * 
	 */
	protected int currentIteration = 0;

	/*
	 * Repository of data records (containing Geometry and SpeciesRecords). Each
	 * record corresponds to the simulation data from one iteration
	 */
	protected IRecordStore recordStore;

	/*
	 * Index within the SimulationRun objects contained by the VRootModel
	 */
	protected int simID = 0;

	private SimulationRun(){
		timestep = 0;
		loadedSimulationModels = new HashMap<String, SimulationModel>();
		availableSimulationModels = new HashMap<String, SimulationModel>();
	}
	
	public SimulationRun(RecordStoreActionListener listener, int simID) {
		this();
		this.simID = simID;
		recordStore = new RecordStore(this);
		recordStore.addListener(listener);
	}

	public IRecordStore getRecordStore() {
		return recordStore;
	}

	/*
	 * Returns the number of records or, if there is no record available, -1
	 */
	public int getIteration() {
		if (recordStore != null)
			return recordStore.size() -1;
		return -1;
	}

	/*
	 * 
	 */
	public int getSimID() {
		return simID;
	}

	/*
	 * 
	 */
	public void setSimID(int simID) {
		this.simID = simID;
	}

	/*
	 * 
	 */
	public Record setIterationUpdated(Record r) {
		Record updatedRecord = recordStore.updateRecord(r);
		return updatedRecord;
	}

	/*
	 * 
	 */
	public int setIncIteration(Record r) {
		recordStore.addRecord(r);

		return recordStore.size();
	}

	public void setIteration(int iteration) {
		if (iteration < recordStore.size())
			recordStore.removeRecordsStartingFrom(iteration);
		else
			while (recordStore.size() <= iteration) {				
				recordStore.addRecord(new Record(recordStore));
			}
	}

	public void setInitialRecord(Record r) {	
		r.setIteration(0);
		if (recordStore.size() == 0)
			recordStore.addRecord(r);
		else
			recordStore.updateRecord(r);
	}

	public void setMaxIterations(int no) {
		maxIterations = no;
	}

	public int getMaxIterations() {
		return maxIterations;
	}

	public Map<String, SimulationModel> getLoadedSimulationModels() {
		return loadedSimulationModels;
	}

	public double getTimestep() {
		return timestep;
	}

	public void setTimestep(double t) {
		timestep = t;
	}

	public SpeciesRecord getSpeciesRecord(Species species, int iteration) {
		return recordStore.getSpeciesRecord(species, iteration);
	}

	public void addAvailableSimulationModel(SimulationModel sm) {
		availableSimulationModels.put(sm.getName(), sm);
	}

	public void clearAvailableSimulationModels() {
		availableSimulationModels.clear();
	}

	public void clearRecordStore() {
		recordStore.clear();
	}

	public void setAvalableSimulationModels(Map<String, SimulationModel> availableModels){
		this.availableSimulationModels = availableModels;
	}
	
	@Override
	public String[] getAvailableModelNames() {
		return availableSimulationModels.keySet().toArray(new String[] {});
	}

	@Override
	public Map<String, Parameter> getModelDefaultParameters(String modelName) {
		return availableSimulationModels.get(modelName).getParameters();
	}

	@Override
	public SimulationModel getAvailableSimulationModel(String modelName) {
		return availableSimulationModels.get(modelName);
	}

	public Map<String, SimulationModel> getAvailableSimulationModels(){
		return availableSimulationModels;				
	}
	
	@Override
	public List<Species> getModelSpecies(String modelName) {
		return loadedSimulationModels.get(modelName).getSpecies();
	}

	public void addLoadedSimulationModel(SimulationModel sm){
		loadedSimulationModels.put(sm.getName(), sm);
	}
	
	public void removeLoadedSimulationModel(String modelName){
		loadedSimulationModels.remove(modelName);
	}
	
	@Override
	public SimulationModel getLoadedSimulationModel(String modelName) {
		return loadedSimulationModels.get(modelName);
	}

	@Override
	public String[] getLoadedSimulationModelNames() {
		return loadedSimulationModels.keySet().toArray(new String[] {});
	}

	@Override
	public Map<String, Species> getLoadedSpecies() {
		Map<String, Species> speciesMap = new HashMap<String, Species>();
		for (String model : loadedSimulationModels.keySet()) {
			SimulationModel sm = loadedSimulationModels.get(model);
			for (Species species : sm.getSpecies())
				if (!species.getName().equals("auxin_wall"))
				speciesMap.put(species.getName(), species);
		}

		return speciesMap;
	}

	@Override
	public Species[] getLoadedSpeciesArray() {
		return getLoadedSpecies().values().toArray(new Species[] {});
	}

	@Override
	public Parameter getLoadedParameter(String modelName, String parameter) {
		return loadedSimulationModels.get(modelName).getParameters()
				.get(parameter);
	}

	@Override
	public Record getLastAvailableRecord() {
		for (int i = recordStore.size() - 1; i >= 0; i--) {
			Record r = recordStore.getRecord(i);
			if (r.getGeometry() != null)
				return r;
		}
		return null;
	}

	@Override
	public Record getRecord() {
		if (recordStore.size() > 0)
			return recordStore.getRecord(recordStore.size() - 1);
		return null;
	}

	@Override
	public Record getRecord(int iteration) {
		if (iteration < recordStore.size())
			return recordStore.getRecord(iteration);
		return null;
	}

	public void setTissueFileInitialized(boolean b) {
		tissueFileInitialised = b;
	}

	public boolean isTissueFileInitialized() {
		return tissueFileInitialised;
	}

	public void setSimulationInitialized(boolean b) {
		simulationInitialised = b;
	}

	public boolean isSimulationInitialized() {
		return simulationInitialised;
	}

	public void setSimInProgress(boolean b) {
		simInProgress = b;
	}

	public boolean isSimInProgress() {
		return simInProgress;
	}

	public String getTissueFileName() {
		return tissueFileName;
	}

	public void setTissueFileName(String fileName) {
		tissueFileName = fileName;
	}

	/*
	 * Registers a RecordStoreActionListener, for the inner recordStore
	 */
	public void addRecordStoreActionListener(RecordStoreActionListener l) {
		recordStore.addListener(l);
	}

	@Override
	public SimulationRun getSimulationRun() {
		// TODO Auto-generated method stub
		return null;
	}

	public void clear() {
		clearRecordStore();
		loadedSimulationModels.clear();
		tissueFileInitialised = false;
		tissueFileName = null;
		simulationInitialised = false;
		currentIteration= 0;
		maxIterations = 0;
		timestep = 0.2;
	}

}
