package rootgui.model.objects;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import rootgui.model.exceptions.ImproperlyStructuredCellWallsException;

/**
 * Store all the information for a cell
 * 
 */
public class Cell implements Serializable {

	/**
	 * 
	 */
	private static final long serialVersionUID = -8009757216165287783L;

	/**
	 * unique id for this cell
	 */
	private int id;

	/**
	 * id of the group that this cell belongs to
	 */
	private int groupId;

	/**
	 * if this cell is truncated
	 */
	private boolean truncated;

	/**
	 * reference to walls of this cell
	 */
	private List<Wall> walls;

	/**
	 * number of points to form the contour of this cell (i.e. number of points
	 * in walls minus duplicates)
	 */
	private int npoints = 0;

	public Cell(int id, int group, List<Wall> walls) {
		this.id = id;
		this.groupId = group;
		this.truncated = false;
		this.walls = walls;
		checkCircuit();
	}

	public Cell(Cell cell) {
		this.id = cell.id;
		this.groupId = cell.groupId;
		// this.concentration=cell.concentration;
		this.truncated = cell.truncated;
		List<Wall> newWalls = new ArrayList<Wall>(walls.size());
		for (int i = 0; i < newWalls.size(); i++) {
			newWalls.set(i, new Wall(this.walls.get(i)));
		}
		this.walls = newWalls;
	}

	public int getId() {
		return id;
	}

	public void setGroup(int group) {
		this.groupId = group;
	}

	public int getGroup() {
		return groupId;
	}

	public void setTruncated(boolean truncated) {
		this.truncated = truncated;
	}

	public List<Wall> getWalls() {
		return walls;
	}

	
	/**
	 * @return list of points to form border (contour) of this cell
	 */
	public List<Point> getBorderPoints() {
		LinkedList<Point> border = new LinkedList<Point>();
		for (Wall w : walls) {
			List<Point> points = w.getPoints();
			assert border.isEmpty() || border.getLast().equals(points.get(0));
			if (!border.isEmpty()) {
				border.removeLast();
			}

			border.addAll(points);
		}
		assert border.getFirst().equals(border.getLast());
		return border;
	}

	
	/**
	 * @return x coordinates of the border points
	 */
	public int[] getXpoints() {
		List<Point> points = getBorderPoints();
		int length = points.size();
		int[] xpoints = new int[length];
		// int[] ypoints = new int[length];
		for (int i = 0; i < length; i++) {
			Point point = points.get(i);
			xpoints[i] = point.getX();
			// ypoints[i] = point.getY();
		}
		npoints = xpoints.length;
		return xpoints;
	}

	/**
	 * @return y coordinates of the border points
	 */
	public int[] getYpoints() {
		List<Point> points = getBorderPoints();
		int length = points.size();
		// int[] xpoints = new int[length];
		int[] ypoints = new int[length];
		for (int i = 0; i < length; i++) {
			Point point = points.get(i);
			// xpoints[i] = point.getX();
			ypoints[i] = point.getY();
		}
		return ypoints;
	}

	public int sizeOfPoints() {
		return npoints;
	}

	@Override
	public boolean equals(Object obj) {
		if (obj instanceof Cell) {
			Cell o = (Cell) obj;
			return id == o.id;
		} else {
			return false;
		}
	}

	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append(String.format("<cell: id=%d, group=%d%n", id, groupId));
		for (Wall w : walls) {
			sb.append("\t");
			sb.append(w.toString());
			sb.append("\n");
		}
		sb.append("/cell>\n");
		return sb.toString();
	}

	/**
	 * Check if the list of walls form a closed circle
	 * @throws ImproperlyStructuredCellWallsException
	 */
	private void checkCircuit() throws ImproperlyStructuredCellWallsException {
		Point head = null, tempTail = null;
		for (Wall w : walls) {
			if (head == null) {
				head = w.getHead();
			}
			if (tempTail != null && !tempTail.equals(w.getHead())) {
				throw new ImproperlyStructuredCellWallsException(this,
						"Walls of the cell fail to form a closed chain");
			}
			tempTail = w.getTail();
		}
		if (!tempTail.equals(head)) {
			throw new ImproperlyStructuredCellWallsException(this,
					"Walls of the cell fail to form a closed chain");
		}
	}

	// /*
	// * Hardcoded enumeration of cell groups
	// */
	// public enum Group {
	// Epiderm(2), Cortex(4), Endoderm(5), Vasculature(5), LRC1(6), LRC2(7),
	// LRC3(
	// 8), LRC4(9), QC(17), Initials(10), S1(11), S2(12), S3(13), S4(14),
	// S5(15);
	//
	// int id;
	//
	// Group(int _id) {
	// this.id = _id;
	// }
	//
	// public int intValue(){
	// return id;
	// }
	//
	// public final static Group[] Columella = new Group[]{Initials, S1, S2, S3,
	// S4, S5};
	// public final static Group[] LRC = new Group[]{LRC1, LRC2, LRC3, LRC4};
	// }

	// /*
	// * Enumeration of cell group zones
	// */
	// public enum Zone {
	// ALL(0), ELONGATING(1), MERISTEMATIC(2);
	//
	// int id;
	//
	// Zone(int _id) {
	// this.id = _id;
	// }
	//
	// public int intValue(){
	// return id;
	// }
	// }

}
