package simserver;

import java.io.IOException;
import java.net.UnknownHostException;
import java.util.Map;
import java.util.Set;

import rootgui.model.Geometry;
import rootgui.model.Parameter;
import rootgui.model.objects.SimulationModel;
import rootgui.model.objects.Species;
import rootgui.model.recordstore.Record;

public interface ISimServerClient {
	/* communication protocol */
	public static final String NUM_ITERATIONS = "iterations";
	public static final String TIMESTEP = "timestep";
	
	/*
	 * Returns true if the client is actively connected to a server socket
	 */
	public boolean isConnected();

	/*
	 * Disconnects from the server
	 */
	public void disconnect();
	
	/*
	 * Causes the ISimServerClient object to 1) disconnect, if it was connected and 2) to create a connection to the server
	 */
	public void reconnect() throws UnknownHostException, IOException;

	/* MUTATORS */

	/*
	 * Specifies the tissue file name and path. Expects a boolean response from
	 * the server.
	 */
	public boolean setTissueFile(String tissueFile);

	/*
	 * Activates (loads) a model in the simulator. Expects a boolean response
	 * from the server.
	 */
	public void addModelByName(String modelName);

	public boolean addModel(SimulationModel sm);

	public boolean addEditedModel(SimulationModel editedModel,
			SimulationModel defaultModel);

	/*
	 * 
	 */
	public boolean setParameter(String modelName, String parameter, String value);

	/*
	 * Specifies to the simulator server, how many iterations it will need to do
	 */
	public int setIterations(int iterations);

	public void setCurrentIteration(int newIteration);

	/*
	 * Sets the timestep of the simulation
	 */
	public double setTimestep(double timestep);

	/* ACCESSORS */

	public String getHostname();


	/*
	 * Asks for a string list of available model names from the server.
	 */
	public String[] getAvailableModelsNames() throws IOException;

	/*
	 * Asks for a dictionary of default model parameters from the server.
	 */
	public Map<String, Parameter> getModelDefaultParameters(String modelName)
			throws Exception;

	public Map<String, Species> getModelSpecies(String modelName,
			Species.Type speciesType);

	/*
	 * Returns the species registered by a model
	 */
	public String[] getAllModelSpeciesNames(String modelName);

	public SimulationModel getSimulationModel(String modelName)
			throws Exception;

	// public String getModelSpecies(String modelName);

	/*
	 * Returns a String[] containing the names of the species contained in the
	 * simulation
	 */
	public Set<Species> getLoadedSpecies();

//	public String getSpeciesPublicName(Species species);
	/*
	 * Returns a double[] containing the current cell values for a given species
	 */
	// public Map<Integer, Double> getCellSpeciesData(String
	// speciesName);

	/*
	 * Returns a Map of wall species data, where the keys are cell ids, while
	 * values contain another Map, from wall id to species value
	 */
	// public Map<Integer, Double> getWallSpeciesData(String
	// speciesName);

	public void setCellSpeciesValue(String cellID, String speciesName,
			double value);

	/*
	 * Returns a String containing the simulated geometry, in CellSet format
	 */
	public String getGeometry();

	/*
	 * 
	 */
	/*
	 * Returns a Record of data obtained from the server, after the tissue has
	 * been initialised (and irrespective if any iterations have been done)
	 */
	public Record getRecord();

	/*
	 * Returns a Record of data obtained from the server. If the
	 * 
	 * @existingGeometry parameter is not null, then the server will be asked if
	 * the geometry has changed, and if it hasn't, it will reuse the
	 * existingGeometry. In all other cases, it will load the geometry from the
	 * server.
	 */
	public Record getRecord(Geometry existingGeometry, Species targetSpecies);

	/*
	 * Returns the current iteration of the client simulation
	 */
	public int getCurrentIteration();

	/*
	 * 
	 */
	public double getTimestep();

	/*
	 * 
	 */
	public String getParameter(String modelName, String parameter);

	/* ACTIVATORS */

	/*
	 * 
	 */
	public void reset();

	/*
	 * 
	 */
	// public int initSimulation();

	/*
	 * Performs another iteration
	 */
	public int step();

	/*
	 * 
	 */
	// public void step(int steps);

	/*
	 * 
	 */
	public boolean initTissue();
	
	public boolean initSimulation();
	
	public boolean setGrowth(boolean onOff);

	public String[] getModelCellSpeciesNames(String modelName);

	public boolean geometryChanged();
}
