package visualiser.viewpanel.view;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.font.FontRenderContext;
import java.awt.font.TextLayout;
import java.awt.geom.AffineTransform;
import java.awt.geom.Path2D;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.SortedMap;
import java.util.TreeMap;

import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.JLabel;
import javax.swing.JPanel;

import rootgui.model.Geometry;
import rootgui.model.objects.Cell;
import rootgui.model.objects.Species;
import rootgui.model.recordstore.Record;
import rootgui.model.recordstore.SpeciesRecord;

import visualiser.Figure; //import visualiser.view.CellView;
//import visualiser.view.CellView;
import visualiser.viewpanel.controller.CDPController;
import visualiser.viewpanel.view.cellrendering.AUX1CellRenderer;
import visualiser.viewpanel.view.cellrendering.AbsCellRenderer;
import visualiser.viewpanel.view.cellrendering.BackgroundCellRenderer;
import visualiser.viewpanel.view.cellrendering.GeometryCellRenderer;
import visualiser.viewpanel.view.cellrendering.LAXCellRenderer;
import visualiser.viewpanel.view.cellrendering.PINCellRenderer;

public class CellDrawingPanel extends JPanel {
	/*
	 * Reference to the controller of the CellDrawingPanel
	 */
	protected CDPController controller;
	
	private final static int COLOUR_BAR_XOFFSET = 100;
	private final static double PREFFERED_EXTENT = 0.9;

	/*
	 * Collection of CellView objects representing the cells of the Geometry
	 */
	protected Collection<CellView> cellsView;

	/*
	 * The Figure object containing information about the species to be rendered
	 * within the VisualiserTab
	 */
	protected Figure figure;

	/*
	 * The index of the species (from the Figure object) that will be rendered
	 */
	protected int figureIndex;

	/*
	 * A reference towards the Species that is being rendered within the
	 * CellDrawingPanel (resulting from figure.getSpecies(figureIndex))
	 */
	protected Species selectedSpecies;

	/*
	 * Boolean flag specifying if the range of the species rendered in the
	 * CellDrawingPanel should be calculated dynamically, or it can be taken
	 * from the figure object (through figure.getMaxRange(figureIndex))
	 */
	protected boolean dynamicMinMaxCalculation = true;

	/*
	 * A collection of cell renderers used for drawing the tissue (and its
	 * species) on the CellDrawingPanel
	 */
	protected ArrayList<AbsCellRenderer> cellRenderers;

	/*
	 * The zoom ratio applied when rendering the tissue
	 */
	//protected double zoomRatio = 1;

	private Color backgroundColour = Color.WHITE;
	
	/*
	 * The stored translate transforms;
	 */
//	protected double tx = 0, ty = 0;

	protected AffineTransform at = new AffineTransform(1, 0, 0,	1, 0, 0);

	/*
	 * A Shape object representing the title of the drawn tissue (basically, the
	 * species' name)
	 */
	protected Shape titleShape;

	/*
	 * An image representing the colour bar
	 */
	private ImageIcon RGBcolourBarIcon = new ImageIcon(this.getClass()
			.getResource("colour bar.png"));
	/*
	 * An image represeting the close CellDrawingPanel button
	 */
	private ImageIcon closeIcon = new ImageIcon(this.getClass().getResource(
			"fileclose.png"));
	/*
	 * ?
	 */
	private ImageIcon greenColourBarIcon = new ImageIcon(this.getClass()
			.getResource("green colour bar.png"));
	private int colourBarType = 0;
	private int closeIconX = 0;
	private int closeIconY = 0;

	/*
	 * 
	 */
	// protected Rectangle boundingRectangle = new Rectangle();
	/*
	 * The id (or index) of the CellDrawingPanel within a VisualiserTab
	 */
	private int id = -1;

	private boolean isExporting = false;

	public CellDrawingPanel(int id) {
		this(id, null, 0);
	}

	public CellDrawingPanel(int id, Figure f, int index) {
		this.id = id;
		this.figure = f;
		figureIndex = index;
		cellRenderers = new ArrayList<AbsCellRenderer>();
		addCellRenderer(new BackgroundCellRenderer());
		setBackground(backgroundColour);
		addCellRenderer(new GeometryCellRenderer());
		setBorder(BorderFactory.createBevelBorder(1));
		titleShape = getSpeciesTitleShape(getSpecies());
	}

	/* CELL RENDERERS */
	/*
	 * Adds another cell renderer (in order to implement layered visualisation
	 */
	public void addCellRenderer(AbsCellRenderer acr) {
		acr.setCellDrawingPanel(this);
		cellRenderers.add(acr);
	}

	/*
	 * 
	 */
	public void removeCellRenderer(String name) {
		cellRenderers.remove(name);
	}

	public void resetCellRenderers() {
		cellRenderers.clear();
	}

	public void setController(CDPController controller) {
		this.controller = controller;

	}

	/* TRANSFORMS */

	/*
	 * Applies an affine transform (zoom and translation factors) to the
	 * existing transform. The method causes the CellDrawingPanel to repaint.
	 */
	public void applyTransform(double s, double tx, double ty,
			boolean isChangedBySpinner) {

		double zoomRatio = figure.getZoomRatio();
		double oldRatio = zoomRatio;
		if (!isChangedBySpinner) {
			if (s < 1) {
				zoomRatio = zoomRatio - (1 - s);
			} else if (s > 1) {
				zoomRatio = zoomRatio + (s - 1);
			}
		} else {
//			tx = this.getWidth()/2 * VisualiserPanel.ZOOM_INCREMENT;
//			ty = this.getHeight()/2 * VisualiserPanel.ZOOM_INCREMENT;
			zoomRatio = s;
		}
		// zoomRatio *= s;
		if(zoomRatio < 0.01)
			zoomRatio = 0.01;
		
		double ratio = (1 / oldRatio) * zoomRatio;
		
//		at.setTransform(ratio, 0, 0, ratio, tx, ty);
		
		at.setToIdentity();
		at.translate(-this.getWidth()/2, -this.getHeight()/2);
		at.scale(ratio,ratio);
		at.translate(tx+this.getWidth()/2, ty+this.getHeight()/2);
//	

		setTransform(at);
		
//		this.tx = (this.tx)*ratio +at.getTranslateX();
//		this.ty = (this.ty)*ratio +at.getTranslateY();
//		figure.setTxTy(this.tx,this.ty);
		
		figure.setTxTy(figure.getTxTy()[0]*ratio +at.getTranslateX(),figure.getTxTy()[1]*ratio +at.getTranslateY());
		figure.setZoomRatio(zoomRatio);
		repaint();
	}

	/*
	 * Sets the zoom ratio for rendering the tissue. The method causes the
	 * CellDrawingPanel to repaint.
	 */
	public void setZoomRatio(double s) {
		if (figure != null) {
			figure.setZoomRatio(s);
		}
//		zoomRatio=s;
		at.setTransform(s, 0, 0, s, at.getTranslateX(), at.getTranslateY());
		setTransform(at);
		repaint();
	
	}

	/*
	 * Returns the zoom ratio used in rendering the tissue
	 */
	public double getZoomRatio() {
//		return zoomRatio;
		return figure.getZoomRatio();
	}

	/*
	 * Returns the zoom ratio that needs to be applied in order for the tissue
	 * to fit the CDP
	 */
	public double getFitZoomRatio() {
		Record currentRecord = controller.getRecord();
		
		if (currentRecord != null) {
		
		Geometry g = currentRecord.getGeometry();
		double z1 = (double) g.getTissueHeight() / getSize().getHeight();
		double z2 = (double) g.getTissueWidth() / (getSize().getWidth() - COLOUR_BAR_XOFFSET - 10);
		double z3 = 0;
		z3 = Math.min(PREFFERED_EXTENT/z1, PREFFERED_EXTENT/z2);
		//minimum scale of 0.01
		if (z3 <=0.01)
			z3 = 0.01;
		
		if (z3 <=0.2)
			z3 = Math.floor(z3 * 100) / 100;
		
		if (z3 >0.2)
			z3 = Math.floor(z3 * 10) / 10;
		
		return z3; 
		}
		return 1;
	}

	/*
	 * Applies the getFitZoomRatio() to the rendered tissue
	 */
	public void setFitZoomRatio() {
		double zoomRatio = getFitZoomRatio();
		setZoomRatio(zoomRatio);
	}

	/*
	 * 
	 */
	private Shape getSpeciesTitleShape(Species species) {
		if (species == null || species.getPublicName() == null
				|| species.getPublicName().length() == 0)
			return null;
//		fontFormat font = species.getFontType();		
		TextLayout text = new TextLayout(species.getPublicName(), new Font(
				"Helvetica", Font.BOLD, 20), new FontRenderContext(null, false, false));		
		int x = 20;
		int y = 40;
		AffineTransform textAt = new AffineTransform();
		textAt.translate(x, y);

		return text.getOutline(textAt);
	}

	/*
	 * Creates the CellView collection and sends the SpeciesRecord to the Cell
	 * renderers
	 */
	public synchronized void initCellsView(final boolean initial) {
		Record currentRecord = controller.getRecord();
		cellsView = new ArrayList<CellView>();
		
		if (currentRecord != null) {
			Geometry g = currentRecord.getGeometry();
			Collection<Cell> cells = g.getCells();
			for (Cell cell : cells) {
				CellView view = new CellView(cell);
				cellsView.add(view);
			}

			titleShape = getSpeciesTitleShape(getSpecies());

			if (cells != null && cells.size() > 0 && 
			   (figure.isFirstVisualisation() || initial)) {
				setFitZoomRatio();
				figure.setFirstVisualisation(false);
			} 

			Rectangle r = g.getBoundingRectangle(); 
			double zoomRatio = figure.getZoomRatio();
			if (initial) {
				at.setTransform(zoomRatio, 0, 0, zoomRatio, -(r.x*zoomRatio) + (this.getWidth() - COLOUR_BAR_XOFFSET - r.width*zoomRatio)/4, -(r.y*zoomRatio) + (this.getHeight() -  r.height*zoomRatio)/2); //-(r.y*zoomRatio) + (this.getHeight() -  r.height*zoomRatio)/2 
				double tx = -(r.x*zoomRatio) + (this.getWidth() - COLOUR_BAR_XOFFSET - r.width*zoomRatio)/4;
				double ty = -(r.y*zoomRatio) + (this.getHeight() -  r.height*zoomRatio)/2;
				figure.setTxTy(tx,ty);
//			this.tx= -(r.x*zoomRatio) + (this.getWidth() - COLOUR_BAR_XOFFSET - r.width*zoomRatio)/4;
//			this.ty= -(r.y*zoomRatio) + (this.getHeight() -  r.height*zoomRatio)/2;
//			
			}else{
			at.setTransform(zoomRatio, 0,0,zoomRatio,figure.getTxTy()[0],figure.getTxTy()[1]); //-(r.y*zoomRatio) + (this.getHeight() -  r.height*zoomRatio)/2 	
//			at.setTransform(zoomRatio, 0,0,zoomRatio,this.tx,this.ty); //-(r.y*zoomRatio) + (this.getHeight() -  r.height*zoomRatio)/2 
			}
			setTransform(at);

			// tells the CellRenderers that a new iteration (Record) is
			// available
		} else {
			
		}
		for (AbsCellRenderer acr : cellRenderers)
			acr.newIteration(currentRecord);
	}

	public Figure getFigure(){
		return figure;
	}
	/*
	 * Sets the Figure and figureIndex of the CellDrawingPanel (basically it
	 * sets or modifies what species is being rendered). The method causes the
	 * CellDrawingPanel to repaint.
	 */
	public void setFigure(Figure f, int index) {
		figure = f;		
		setFigureIndex(index);
	}

	/*
	 * Returns the index, within the Figure object, of the species that the
	 * CellDrawingPanel is rendering.
	 */
	public int getFigureIndex() {
		return figureIndex;
	}

	/*
	 * Modifies the figure index: the index, within the Figure object, of the
	 * species that the CellDrawingPanel is rendering. The method causes the
	 * CellDrawingPanel to repaint. The method only works with two cell
	 * renderers at the moment: a variable one (index 0) and a
	 * GeometryCellRenderer (index 1)
	 */
	public void setFigureIndex(int index) {												
		dynamicMinMaxCalculation = figure.getMaxRange(figureIndex) <= 0;
		
		boolean newSpecies = false;
		if (selectedSpecies == null) {
			selectedSpecies = figure.getSpecies(index);
			newSpecies = true;
		}
		if (!newSpecies && !figure.getSpecies(index).equals(selectedSpecies)) {
			newSpecies = true; 			
			selectedSpecies = figure.getSpecies(index);
		}		

		if (newSpecies) {
			titleShape = getSpeciesTitleShape(selectedSpecies);
			switch (selectedSpecies.getType()) {
			case CELL:
				if (!(cellRenderers.get(0) instanceof BackgroundCellRenderer))
					cellRenderers.set(0, new BackgroundCellRenderer());
				break;
			case CELL_WALL:
				if (selectedSpecies.getName().equals("PIN")) {
					if (!(cellRenderers.get(0) instanceof PINCellRenderer))
						cellRenderers.set(0, new PINCellRenderer());
				} else if (selectedSpecies.getName().startsWith("AUX")) {
					if (!(cellRenderers.get(0) instanceof AUX1CellRenderer))
						cellRenderers.set(0, new AUX1CellRenderer());
				} else if (selectedSpecies.getName().startsWith("LAX")) {
					if (!(cellRenderers.get(0) instanceof LAXCellRenderer))
						cellRenderers.set(0, new LAXCellRenderer());
				}
				break;
			default:
				break;
			}		
		cellRenderers.get(0).setCellDrawingPanel(this);
		}
		// cellRenderers.add(new GeometryCellRenderer());
		figureIndex = index;
		for (AbsCellRenderer acr : cellRenderers)
			acr.setFigure(figure, index, controller.getRecord());

		repaint();
	}

	/*
	 * Returns the Species being rendered by the CellDrawingPanel
	 */
	public Species getSpecies() {
		if (selectedSpecies != null)
			return selectedSpecies;
		return new Species("", "", null);
	}

	/*
	 * Returns the id (index) of the CellDrawingPanel, within the VisualiserTab
	 */
	public int getId() {
		return id;
	}

	public int getColourBarType() {
		return colourBarType;
	}

	/*
	 * Paints the colour bar
	 */
	private void paintColourBar(double minValue, double maxValue,
			int noDivisions, Graphics2D g2) {
		ImageIcon colourBarIcon;
		
		if (colourBarType == 0)
			colourBarIcon = RGBcolourBarIcon;
		else
			colourBarIcon = greenColourBarIcon;
		int imageWidth = colourBarIcon.getIconWidth();
		int imageHeight = colourBarIcon.getIconHeight();

		int x = (int) (getVisibleRect().getWidth() + getVisibleRect().getX()) - COLOUR_BAR_XOFFSET;
		
		int y = (int) ((getVisibleRect().getHeight() + getVisibleRect().getY()) - imageHeight)/2; 
		
		g2.drawImage(colourBarIcon.getImage(), x, y, null);
		double divH = (double) (imageHeight - 15) / noDivisions;
		double range = maxValue - minValue;
		int lblX = x + imageWidth + 10;
		
		if(getColourBarType()==0){
			g2.setColor(Color.BLACK);
			}
		if(getColourBarType()==1){
			//g2.setColor(Color.WHITE);
			g2.setColor(Color.BLACK);
		}
		
		for (int i = noDivisions; i >= 0; i--) {
			double d = minValue + (range * (1 - (double) i / noDivisions));
			String s = String.format("%.1f", d);
			int lblY = (int) (y + i * divH) + 10;
			g2.setBackground(Color.WHITE);
			g2.drawString(s, lblX, lblY);
		}
	}

	public synchronized void paintComponent(Graphics g) {
		super.paintComponent(g); 
		Graphics2D g2 = (Graphics2D) g;
//		System.err.println("CDP: paintComponent");

		// g2.setTransform(at);
		Record r = controller.getRecord();
		boolean backgroundRenderer = false;
		
		if (r != null && cellsView != null) {
			// drawing the cells
			double maxRange = 0;
			for (AbsCellRenderer acr : cellRenderers) {
				if (acr instanceof BackgroundCellRenderer) {
					backgroundRenderer = true;
					maxRange = ((BackgroundCellRenderer) acr).getMaxRange();
					maxRange = Math.ceil(maxRange);
					((BackgroundCellRenderer) acr).setMaxRange(maxRange);
//					System.out.println("Displayed Max Range = "+maxRange);
				}
				acr.drawCells(cellsView, g);
			}

			if (titleShape != null) {
				
				if(getColourBarType()==0){
					g2.setColor(Color.BLACK);
					}
				if(getColourBarType()==1){
//					g2.setColor(Color.WHITE);
					g2.setColor(Color.BLACK);
				}
				
				g2.fill(titleShape);
				if (backgroundRenderer) {
					paintColourBar(0, Math.ceil(maxRange), 9, g2);
				}
			}
						
			AbsCellRenderer.highlightCells(controller.getSelectedCellViews(), g);
			if (!isExporting) {
				closeIconX = (int) (getVisibleRect().getWidth() + getVisibleRect().getX()) - 10 - closeIcon.getIconWidth();
				closeIconY = (int) getVisibleRect().getY() + 10;
				g2.drawImage(closeIcon.getImage(), closeIconX,
								closeIconY, null);
			}
//			g2.draw(r.getGeometry().getBoundingRectangle());
		}
	}

	public CellView pointInCell(Point p) {
		for (CellView view : cellsView) {
			if (view.contains(p)) {
				// view.setSelected(true);
				return view;
			}
		}
		return null;
	}

	public void setExporting(boolean b) {
		isExporting = b;
	}

	/*
	 * Returns true if the mouse was clicked over the Close icon
	 */
	public boolean isPressedCloseIcon(Point p) {
		double x = p.getX();
		double y = p.getY();
		// 22 is the icon's height and width
		if (x >= closeIconX && x <= closeIconX + 22 && y >= closeIconY
				&& y <= closeIconY + 22)
			return true;
		return false;
	}

	public void setColourBar(int i) {
		colourBarType = i;
	}

	public CellView getCellViewById(int cellId) {
		for (CellView view : cellsView) {
			if (view.getCellID() == cellId) {
				// view.setSelected(true);
				return view;
			}
		}
		return null;
	}

	private void setTransform(AffineTransform at) {
		this.at = at;
		if (cellsView != null) {
			for (CellView view : cellsView) {
				view.transform(at);
			}			
		}
	}
}
